@@ -1,4 +1,4 @@
-{-# LANGUAGE BangPatterns, ScopedTypeVariables #-}
+{-# LANGUAGE BangPatterns, ScopedTypeVariables, ForeignFunctionInterface #-}
 -- |
 -- Module    : Numeric.SpecFunctions.Internal
 -- Copyright : (c) 2009, 2011, 2012 Bryan O'Sullivan
@@ -11,16 +11,20 @@
 -- Internal module with implementation of special functions.
 module Numeric.SpecFunctions.Internal where
 
+import Control.Applicative
 import Data.Bits       ((.&.), (.|.), shiftR)
 import Data.Int        (Int64)
-import qualified Data.Number.Erf     as Erf (erfc,erf)
+import Data.Word       (Word)
 import qualified Data.Vector.Unboxed as U
+import           Data.Vector.Unboxed   ((!))
 
 import Numeric.Polynomial.Chebyshev    (chebyshevBroucke)
-import Numeric.Polynomial              (evaluateEvenPolynomialL,evaluateOddPolynomialL)
+import Numeric.Polynomial              (evaluatePolynomialL,evaluateEvenPolynomialL,evaluateOddPolynomialL)
+import Numeric.RootFinding             (Root(..), newtonRaphson)
+import Numeric.Series                  (sumPowerSeries,enumSequenceFrom,scanSequence,evalContFractionB)
 import Numeric.MathFunctions.Constants ( m_epsilon, m_NaN, m_neg_inf, m_pos_inf
-                                       , m_sqrt_2_pi, m_ln_sqrt_2_pi, m_sqrt_2
-                                       , m_eulerMascheroni
+                                       , m_sqrt_2_pi, m_ln_sqrt_2_pi, m_eulerMascheroni
+                                       , m_min_log, m_tiny
                                        )
 import Text.Printf
 
@@ -31,21 +35,44 @@ import Text.Printf
 
 -- | Error function.
 --
--- > erf -∞ = -1
--- > erf  0 =  0
--- > erf +∞ =  1
+-- \[
+-- \operatorname{erf}(x) = \frac{2}{\sqrt{\pi}} \int_{0}^{x} \exp(-t^2) dt
+-- \]
+--
+-- Function limits are:
+--
+-- \[
+-- \begin{aligned}
+--  &\operatorname{erf}(-\infty) &=& -1 \\
+--  &\operatorname{erf}(0)       &=& \phantom{-}\,0 \\
+--  &\operatorname{erf}(+\infty) &=& \phantom{-}\,1 \\
+-- \end{aligned}
+-- \]
 erf :: Double -> Double
 {-# INLINE erf #-}
-erf = Erf.erf
+erf = c_erf
 
 -- | Complementary error function.
 --
--- > erfc -∞ = 2
--- > erfc  0 = 1
--- > errc +∞ = 0
+-- \[
+-- \operatorname{erfc}(x) = 1 - \operatorname{erf}(x)
+-- \]
+--
+-- Function limits are:
+--
+-- \[
+-- \begin{aligned}
+--  &\operatorname{erf}(-\infty) &=&\, 2 \\
+--  &\operatorname{erf}(0)       &=&\, 1 \\
+--  &\operatorname{erf}(+\infty) &=&\, 0 \\
+-- \end{aligned}
+-- \]
 erfc :: Double -> Double
 {-# INLINE erfc #-}
-erfc = Erf.erfc
+erfc = c_erfc
+
+foreign import ccall "erf"  c_erf  :: Double -> Double
+foreign import ccall "erfc" c_erfc :: Double -> Double
 
 
 -- | Inverse of 'erf'.
@@ -81,103 +108,59 @@ invErfc p
 -- Gamma function
 ----------------------------------------------------------------
 
--- Adapted from http://people.sc.fsu.edu/~burkardt/f_src/asa245/asa245.html
-
--- | Compute the logarithm of the gamma function Γ(/x/).  Uses
--- Algorithm AS 245 by Macleod.
---
--- Gives an accuracy of 10-12 significant decimal digits, except
--- for small regions around /x/ = 1 and /x/ = 2, where the function
--- goes to zero.  For greater accuracy, use 'logGammaL'.
---
--- Returns ∞ if the input is outside of the range (0 < /x/ ≤ 1e305).
-logGamma :: Double -> Double
-logGamma x
-    | x <= 0    = m_pos_inf
-    -- Handle positive infinity. logGamma overflows before 1e308 so
-    -- it's safe
-    | x > 1e308 = m_pos_inf
-    -- Normal cases
-    | x < 1.5   = a + c *
-                  ((((r1_4 * b + r1_3) * b + r1_2) * b + r1_1) * b + r1_0) /
-                  ((((b + r1_8) * b + r1_7) * b + r1_6) * b + r1_5)
-    | x < 4     = (x - 2) *
-                  ((((r2_4 * x + r2_3) * x + r2_2) * x + r2_1) * x + r2_0) /
-                  ((((x + r2_8) * x + r2_7) * x + r2_6) * x + r2_5)
-    | x < 12    = ((((r3_4 * x + r3_3) * x + r3_2) * x + r3_1) * x + r3_0) /
-                  ((((x + r3_8) * x + r3_7) * x + r3_6) * x + r3_5)
-    | x > 3e6   = k
-    | otherwise = k + x1 *
-                  ((r4_2 * x2 + r4_1) * x2 + r4_0) /
-                  ((x2 + r4_4) * x2 + r4_3)
-  where
-    (a , b , c)
-        | x < 0.5   = (-y , x + 1 , x)
-        | otherwise = (0  , x     , x - 1)
-
-    y      = log x
-    k      = x * (y-1) - 0.5 * y + alr2pi
-    alr2pi = 0.918938533204673
-
-    x1 = 1 / x
-    x2 = x1 * x1
-
-    r1_0 =  -2.66685511495;   r1_1 =  -24.4387534237;    r1_2 = -21.9698958928
-    r1_3 =  11.1667541262;    r1_4 =    3.13060547623;   r1_5 =   0.607771387771
-    r1_6 =  11.9400905721;    r1_7 =   31.4690115749;    r1_8 =  15.2346874070
-
-    r2_0 = -78.3359299449;    r2_1 = -142.046296688;     r2_2 = 137.519416416
-    r2_3 =  78.6994924154;    r2_4 =    4.16438922228;   r2_5 =  47.0668766060
-    r2_6 = 313.399215894;     r2_7 =  263.505074721;     r2_8 =  43.3400022514
-
-    r3_0 =  -2.12159572323e5; r3_1 =    2.30661510616e5; r3_2 =   2.74647644705e4
-    r3_3 =  -4.02621119975e4; r3_4 =   -2.29660729780e3; r3_5 =  -1.16328495004e5
-    r3_6 =  -1.46025937511e5; r3_7 =   -2.42357409629e4; r3_8 =  -5.70691009324e2
-
-    r4_0 = 0.279195317918525;  r4_1 = 0.4917317610505968;
-    r4_2 = 0.0692910599291889; r4_3 = 3.350343815022304
-    r4_4 = 6.012459259764103
-
-
-
 data L = L {-# UNPACK #-} !Double {-# UNPACK #-} !Double
 
--- | Compute the logarithm of the gamma function, &#915;(/x/).  Uses a
--- Lanczos approximation.
+-- | Compute the logarithm of the gamma function, Γ(/x/).
 --
--- This function is slower than 'logGamma', but gives 14 or more
--- significant decimal digits of accuracy, except around /x/ = 1 and
--- /x/ = 2, where the function goes to zero.
+-- \[
+-- \Gamma(x) = \int_0^{\infty}t^{x-1}e^{-t}\,dt = (x - 1)!
+-- \]
+--
+-- This implementation uses Lanczos approximation. It gives 14 or more
+-- significant decimal digits, except around /x/ = 1 and /x/ = 2,
+-- where the function goes to zero.
 --
 -- Returns &#8734; if the input is outside of the range (0 < /x/
 -- &#8804; 1e305).
+logGamma :: Double -> Double
+logGamma x
+  | x <= 0    = m_pos_inf
+  | x <  1    = lanczos (1+x) - log x
+  | otherwise = lanczos x
+  where
+    -- Evaluate Lanczos approximation for γ=6
+    lanczos z = fini
+              $ U.foldl' go (L 0 (z+7)) a
+      where
+        fini (L l _)   = log (l+a0) + log m_sqrt_2_pi - z65 + (z-0.5) * log z65
+        go   (L l t) k = L (l + k / t) (t-1)
+        z65 = z + 6.5
+    -- Coefficients for Lanczos approximation
+    a0  = 0.9999999999995183
+    a   = U.fromList [ 0.1659470187408462e-06
+                     , 0.9934937113930748e-05
+                     , -0.1385710331296526
+                     , 12.50734324009056
+                     , -176.6150291498386
+                     , 771.3234287757674
+                     , -1259.139216722289
+                     , 676.5203681218835
+                     ]
+
+-- | Synonym for 'logGamma'. Retained for compatibility
 logGammaL :: Double -> Double
-logGammaL x
-    | x <= 0    = m_pos_inf
-    -- Lanroz approximation loses precision for small arguments
-    | x <= 1e-3 = logGamma x
-    | otherwise = fini . U.foldl' go (L 0 (x+7)) $ a
-    where fini (L l _) = log (l+a0) + log m_sqrt_2_pi - x65 + (x-0.5) * log x65
-          go (L l t) k = L (l + k / t) (t-1)
-          x65 = x + 6.5
-          a0  = 0.9999999999995183
-          a   = U.fromList [ 0.1659470187408462e-06
-                           , 0.9934937113930748e-05
-                           , -0.1385710331296526
-                           , 12.50734324009056
-                           , -176.6150291498386
-                           , 771.3234287757674
-                           , -1259.139216722289
-                           , 676.5203681218835
-                           ]
+logGammaL = logGamma
 
 
-
--- | Compute the log gamma correction factor for @x@ &#8805; 10.  This
--- correction factor is suitable for an alternate (but less
--- numerically accurate) definition of 'logGamma':
+-- |
+-- Compute the log gamma correction factor for Stirling
+-- approximation for @x@ &#8805; 10.  This correction factor is
+-- suitable for an alternate (but less numerically accurate)
+-- definition of 'logGamma':
 --
--- >lgg x = 0.5 * log(2*pi) + (x-0.5) * log x - x + logGammaCorrection x
+-- \[
+-- \log\Gamma(x) = \frac{1}{2}\log(2\pi) + (x-\frac{1}{2})\log x - x + \operatorname{logGammaCorrection}(x)
+-- \]
 logGammaCorrection :: Double -> Double
 logGammaCorrection x
     | x < 10    = m_NaN
@@ -199,83 +182,109 @@ logGammaCorrection x
 
 
 -- | Compute the normalized lower incomplete gamma function
--- γ(/s/,/x/). Normalization means that
--- γ(/s/,∞)=1. Uses Algorithm AS 239 by Shea.
-incompleteGamma :: Double       -- ^ /s/ ∈ (0,∞)
+-- γ(/z/,/x/). Normalization means that γ(/z/,∞)=1
+--
+-- \[
+-- \gamma(z,x) = \frac{1}{\Gamma(z)}\int_0^{x}t^{z-1}e^{-t}\,dt
+-- \]
+--
+-- Uses Algorithm AS 239 by Shea.
+incompleteGamma :: Double       -- ^ /z/ ∈ (0,∞)
                 -> Double       -- ^ /x/ ∈ (0,∞)
                 -> Double
-incompleteGamma p x
-    | isNaN p || isNaN x = m_NaN
-    | x < 0 || p <= 0    = m_pos_inf
-    | x == 0             = 0
-    -- For very large `p' normal approximation gives <1e-10 error
-    | p >= 2e5           = norm (3 * sqrt p * ((x/p) ** (1/3) + 1/(9*p) - 1))
-    | p >= 500           = approx
-    -- Dubious approximation
-    | x >= 1e8           = 1
-    | x <= 1 || x < p    = let a = p * log x - x - logGamma (p + 1)
-                               g = a + log (pearson p 1 1)
-                           in if g > limit then exp g else 0
-    | otherwise          = let g = p * log x - x - logGamma p + log cf
-                           in if g > limit then 1 - exp g else 1
+-- Notation used:
+--  + P(a,x) - regularized lower incomplete gamma
+--  + Q(a,x) - regularized upper incomplete gamma
+incompleteGamma a x
+  | a <= 0 || x < 0 = error
+     $ "incompleteGamma: Domain error z=" ++ show a ++ " x=" ++ show x
+  | x == 0          = 0
+  | x == m_pos_inf  = 1
+  -- For very small x we use following expansion for P:
+  --
+  -- See http://functions.wolfram.com/GammaBetaErf/GammaRegularized/06/01/05/01/01/
+  | x < sqrt m_epsilon && a > 1
+    = x**a / a / exp (logGammaL a) * (1 - a*x / (a + 1))
+  | x < 0.5 = case () of
+    _| (-0.4)/log x < a  -> taylorSeriesP
+     | otherwise         -> taylorSeriesComplQ
+  | x < 1.1 = case () of
+    _| 0.75*x < a        -> taylorSeriesP
+     | otherwise         -> taylorSeriesComplQ
+  | a > 20 && useTemme    = uniformExpansion
+  | x - (1 / (3 * x)) < a = taylorSeriesP
+  | otherwise             = contFraction
   where
-    -- CDF for standard normal distributions
-    norm a = 0.5 * erfc (- a / m_sqrt_2)
-    -- For large values of `p' we use 18-point Gauss-Legendre
-    -- integration.
-    approx
-      | ans > 0   = 1 - ans
-      | otherwise = -ans
-      where
-        -- Set upper limit for integration
-        xu | x > p1    =         (p1 + 11.5*sqrtP1) `max` (x + 6*sqrtP1)
-           | otherwise = max 0 $ (p1 -  7.5*sqrtP1) `min` (x - 5*sqrtP1)
-        s = U.sum $ U.zipWith go coefY coefW
-        go y w = let t = x + (xu - x)*y
-                 in w * exp( -(t-p1) + p1*(log t - lnP1) )
-        ans = s * (xu - x) * exp( p1 * (lnP1 - 1) - logGamma p)
-        --
-        p1     = p - 1
-        lnP1   = log  p1
-        sqrtP1 = sqrt p1
+    mu = (x - a) / a
+    useTemme = (a > 200 && 20/a > mu*mu)
+            || (abs mu < 0.4)
+    -- Gautschi's algorithm.
     --
-    pearson !a !c !g
-        | c' <= tolerance = g'
-        | otherwise       = pearson a' c' g'
-        where a' = a + 1
-              c' = c * x / a'
-              g' = g + c'
-    cf = let a = 1 - p
-             b = a + x + 1
-             p3 = x + 1
-             p4 = x * b
-         in contFrac a b 0 1 x p3 p4 (p3/p4)
-    contFrac !a !b !c !p1 !p2 !p3 !p4 !g
-        | abs (g - rn) <= min tolerance (tolerance * rn) = g
-        | otherwise = contFrac a' b' c' (f p3) (f p4) (f p5) (f p6) rn
-        where a' = a + 1
-              b' = b + 2
-              c' = c + 1
-              an = a' * c'
-              p5 = b' * p3 - an * p1
-              p6 = b' * p4 - an * p2
-              rn = p5 / p6
-              f n | abs p5 > overflow = n / overflow
-                  | otherwise         = n
-    limit     = -88
-    tolerance = 1e-14
-    overflow  = 1e37
+    -- Evaluate series for P(a,x). See [Temme1994] Eq. 5.5
+    --
+    -- FIXME: Term `exp (log x * z - x - logGamma (z+1))` doesn't give full precision
+    taylorSeriesP
+      = sumPowerSeries x (scanSequence (/) 1 $ enumSequenceFrom (a+1))
+      * exp (log x * a - x - logGamma (a+1))
+    -- Series for 1-Q(a,x). See [Temme1994] Eq. 5.5
+    taylorSeriesComplQ
+      = sumPowerSeries (-x) (scanSequence (/) 1 (enumSequenceFrom 1) / enumSequenceFrom a)
+      * x**a / exp(logGammaL a)
+    -- Legendre continued fractions
+    contFraction = 1 - ( exp ( log x * a - x - logGamma a )
+                       / evalContFractionB frac
+                       )
+      where
+        frac = (\k -> (k*(a-k), x - a + 2*k + 1)) <$> enumSequenceFrom 0
+    -- Evaluation based on uniform expansions. See [Temme1994] 5.2
+    uniformExpansion =
+      let -- Coefficients f_m in paper
+          fm :: U.Vector Double
+          fm = U.fromList [ 1.00000000000000000000e+00
+                          ,-3.33333333333333370341e-01
+                          , 8.33333333333333287074e-02
+                          ,-1.48148148148148153802e-02
+                          , 1.15740740740740734316e-03
+                          , 3.52733686067019369930e-04
+                          ,-1.78755144032921825352e-04
+                          , 3.91926317852243766954e-05
+                          ,-2.18544851067999240532e-06
+                          ,-1.85406221071515996597e-06
+                          , 8.29671134095308545622e-07
+                          ,-1.76659527368260808474e-07
+                          , 6.70785354340149841119e-09
+                          , 1.02618097842403069078e-08
+                          ,-4.38203601845335376897e-09
+                          , 9.14769958223679020897e-10
+                          ,-2.55141939949462514346e-11
+                          ,-5.83077213255042560744e-11
+                          , 2.43619480206674150369e-11
+                          ,-5.02766928011417632057e-12
+                          , 1.10043920319561347525e-13
+                          , 3.37176326240098513631e-13
+                          ]
+          y   = - log1pmx mu
+          eta = sqrt (2 * y) * signum mu
+          -- Evaluate S_α (Eq. 5.9)
+          loop !_  !_  u 0 = u
+          loop bm1 bm0 u i = let t  = (fm ! i) + (fromIntegral i + 1)*bm1 / a
+                                 u' = eta * u + t
+                             in  loop bm0 t u' (i-1)
+          s_a = let n = U.length fm
+                in  loop (fm ! (n-1)) (fm ! (n-2)) 0 (n-3)
+                  / exp (logGammaCorrection a)
+      in 1/2 * erfc(-eta*sqrt(a/2)) - exp(-(a*y)) / sqrt (2*pi*a) * s_a
 
 
 
 -- Adapted from Numerical Recipes §6.2.1
 
 -- | Inverse incomplete gamma function. It's approximately inverse of
---   'incompleteGamma' for the same /s/. So following equality
+--   'incompleteGamma' for the same /z/. So following equality
 --   approximately holds:
 --
--- > invIncompleteGamma s . incompleteGamma s = id
-invIncompleteGamma :: Double    -- ^ /s/ ∈ (0,∞)
+-- > invIncompleteGamma z . incompleteGamma z ≈ id
+invIncompleteGamma :: Double    -- ^ /z/ ∈ (0,∞)
                    -> Double    -- ^ /p/ ∈ [0,1]
                    -> Double
 invIncompleteGamma a p
@@ -341,13 +350,27 @@ invIncompleteGamma a p
 ----------------------------------------------------------------
 
 -- | Compute the natural logarithm of the beta function.
-logBeta :: Double -> Double -> Double
+--
+-- \[
+-- B(a,b) = \int_0^1 t^{a-1}(1-t)^{1-b}\,dt = \frac{\Gamma(a)\Gamma(b)}{\Gamma(a+b)}
+-- \]
+logBeta
+  :: Double                     -- ^ /a/ > 0
+  -> Double                     -- ^ /b/ > 0
+  -> Double
 logBeta a b
     | p < 0     = m_NaN
     | p == 0    = m_pos_inf
-    | p >= 10   = log q * (-0.5) + m_ln_sqrt_2_pi + logGammaCorrection p + c +
-                  (p - 0.5) * log ppq + q * log1p(-ppq)
-    | q >= 10   = logGamma p + c + p - p * log pq + (q - 0.5) * log1p(-ppq)
+    | p >= 10   = log q * (-0.5)
+                + m_ln_sqrt_2_pi
+                + logGammaCorrection p
+                + c
+                + (p - 0.5) * log ppq + q * log1p(-ppq)
+    | q >= 10   = logGamma p
+                + c
+                + p
+                - p * log pq
+                + (q - 0.5) * log1p(-ppq)
     | otherwise = logGamma p + logGamma q - logGamma pq
     where
       p   = min a b
@@ -356,11 +379,16 @@ logBeta a b
       pq  = p + q
       c   = logGammaCorrection q - logGammaCorrection pq
 
--- | Regularized incomplete beta function. Uses algorithm AS63 by
--- Majumder and Bhattachrjee and quadrature approximation for large
--- /p/ and /q/.
-incompleteBeta :: Double -- ^ /p/ > 0
-               -> Double -- ^ /q/ > 0
+-- | Regularized incomplete beta function.
+--
+-- \[
+-- I(x;a,b) = \frac{1}{B(a,b)} \int_0^x t^{a-1}(1-t)^{1-b}\,dt
+-- \]
+--
+-- Uses algorithm AS63 by Majumder and Bhattachrjee and quadrature
+-- approximation for large /p/ and /q/.
+incompleteBeta :: Double -- ^ /a/ > 0
+               -> Double -- ^ /b/ > 0
                -> Double -- ^ /x/, must lie in [0,1] range
                -> Double
 incompleteBeta p q = incompleteBeta_ (logBeta p q) p q
@@ -368,15 +396,15 @@ incompleteBeta p q = incompleteBeta_ (logBeta p q) p q
 -- | Regularized incomplete beta function. Same as 'incompleteBeta'
 -- but also takes logarithm of beta function as parameter.
 incompleteBeta_ :: Double -- ^ logarithm of beta function for given /p/ and /q/
-                -> Double -- ^ /p/ > 0
-                -> Double -- ^ /q/ > 0
+                -> Double -- ^ /a/ > 0
+                -> Double -- ^ /b/ > 0
                 -> Double -- ^ /x/, must lie in [0,1] range
                 -> Double
 incompleteBeta_ beta p q x
   | p <= 0 || q <= 0            =
       modErr $ printf "incompleteBeta_: p <= 0 || q <= 0. p=%g q=%g x=%g" p q x
   | x <  0 || x >  1 || isNaN x =
-      modErr $ printf "incompletBeta_: x out of [0,1] range. p=%g q=%g x=%g" p q x
+      modErr $ printf "incompleteBeta_: x out of [0,1] range. p=%g q=%g x=%g" p q x
   | x == 0 || x == 1            = x
   | p >= (p+q) * x   = incompleteBetaWorker beta p q x
   | otherwise        = 1 - incompleteBetaWorker beta q p (1 - x)
@@ -394,8 +422,8 @@ incompleteBetaApprox beta p q x
     p1    = p - 1
     q1    = q - 1
     mu    = p / (p + q)
-    lnmu  = log mu
-    lnmuc = log (1 - mu)
+    lnmu  = log     mu
+    lnmuc = log1p (-mu)
     -- Upper limit for integration
     xu = max 0 $ min (mu - 10*t) (x - 5*t)
        where
@@ -419,9 +447,22 @@ incompleteBetaWorker beta p q x
     -- Constants
     eps = 1e-15
     cx  = 1 - x
-    -- Loop
+    -- Common multiplies for expansion. Accurate calculation is a bit
+    -- tricky. Performing calculation in log-domain leads to slight
+    -- loss of precision for small x, while using ** prone to
+    -- underflows.
+    --
+    -- If either beta function of x**p·(1-x)**(q-1) underflows we
+    -- switch to log domain. It could waste work but there's no easy
+    -- switch criterion.
+    factor
+      | beta < m_min_log || prod < m_tiny = exp( p * log x + (q - 1) * log cx - beta)
+      | otherwise                         = prod / exp beta
+      where
+        prod =  x**p * cx**(q - 1)
+    -- Soper's expansion of incomplete beta function
     loop !psq (ns :: Int) ai term betain
-      | done      = betain' * exp( p * log x + (q - 1) * log cx - beta) / p
+      | done      = betain' * factor / p
       | otherwise = loop psq' (ns - 1) (ai + 1) term' betain'
       where
         -- New values
@@ -439,9 +480,9 @@ incompleteBetaWorker beta p q x
 -- | Compute inverse of regularized incomplete beta function. Uses
 -- initial approximation from AS109, AS64 and Halley method to solve
 -- equation.
-invIncompleteBeta :: Double     -- ^ /p/ > 0
-                  -> Double     -- ^ /q/ > 0
-                  -> Double     -- ^ /a/ ∈ [0,1]
+invIncompleteBeta :: Double     -- ^ /a/ > 0
+                  -> Double     -- ^ /b/ > 0
+                  -> Double     -- ^ /x/ ∈ [0,1]
                   -> Double
 invIncompleteBeta p q a
   | p <= 0 || q <= 0 =
@@ -449,13 +490,11 @@ invIncompleteBeta p q a
   | a <  0 || a >  1 =
       modErr $ printf "invIncompleteBeta x must be in [0,1].  p=%g q=%g a=%g" p q a
   | a == 0 || a == 1 = a
-  | a > 0.5          = 1 - invIncompleteBetaWorker (logBeta p q) q p (1 - a)
-  | otherwise        =     invIncompleteBetaWorker (logBeta p q) p q  a
+  | otherwise        = invIncompleteBetaWorker (logBeta p q) p q  a
 
 
 invIncompleteBetaWorker :: Double -> Double -> Double -> Double -> Double
--- NOTE: p <= 0.5.
-invIncompleteBetaWorker beta a b p = loop (0::Int) guess
+invIncompleteBetaWorker beta a b p = loop (0::Int) (invIncBetaGuess beta a b p)
   where
     a1 = a - 1
     b1 = b - 1
@@ -478,64 +517,180 @@ invIncompleteBetaWorker beta a b p = loop (0::Int) guess
       where
         -- Calculate Halley step.
         f   = incompleteBeta_ beta a b x - p
-        f'  = exp $ a1 * log x + b1 * log (1 - x) - beta
+        f'  = exp $ a1 * log x + b1 * log1p (-x) - beta
         u   = f / f'
-        dx  = u / (1 - 0.5 * min 1 (u * (a1 / x - b1 / (1 - x))))
+        -- We bound Halley correction to Newton-Raphson to (-1,1) range
+        corr | d > 1     = 1
+             | d < -1    = -1
+             | isNaN d   = 0
+             | otherwise = d
+          where
+            d = u * (a1 / x - b1 / (1 - x))
+        dx  = u / (1 - 0.5 * corr)
         -- Next approximation. If Halley step leads us out of [0,1]
         -- range we revert to bisection.
         x'  | z < 0     = x / 2
             | z > 1     = (x + 1) / 2
             | otherwise = z
             where z = x - dx
-    -- Calculate initial guess. Approximations from AS64, AS109 and
-    -- Numerical recipes are used.
-    --
-    -- Equations are referred to by name of paper and number e.g. [AS64 2]
-    -- In AS64 papers equations are not numbered so they are refered
-    -- to by number of appearance starting from definition of
-    -- incomplete beta.
-    guess
-      -- In this region we use approximation from AS109 (Carter
-      -- approximation). It's reasonably good (2 iterations on
-      -- average)
-      | a > 1 && b > 1 =
-          let r = (y*y - 3) / 6
-              s = 1 / (2*a - 1)
-              t = 1 / (2*b - 1)
-              h = 2 / (s + t)
-              w = y * sqrt(h + r) / h - (t - s) * (r + 5/6 - 2 / (3 * h))
-          in a / (a + b * exp(2 * w))
-      -- Otherwise we revert to approximation from AS64 derived from
-      -- [AS64 2] when it's applicable.
-      --
-      -- It slightly reduces average number of iterations when `a' and
-      -- `b' have different magnitudes.
-      | chi2 > 0 && ratio > 1 = 1 - 2 / (ratio + 1)
-      -- If all else fails we use approximation from "Numerical
-      -- Recipes". It's very similar to approximations [AS64 4,5] but
-      -- it never goes out of [0,1] interval.
-      | otherwise = case () of
-          _| p < t / w  -> (a * p * w) ** (1/a)
-           | otherwise  -> 1 - (b * (1 - p) * w) ** (1/b)
-           where
-             lna = log $ a / (a+b)
-             lnb = log $ b / (a+b)
-             t   = exp( a * lna ) / a
-             u   = exp( b * lnb ) / b
-             w   = t + u
+
+
+-- Calculate initial guess for inverse incomplete beta function.
+invIncBetaGuess :: Double -> Double -> Double -> Double -> Double
+-- Calculate initial guess. for solving equation for inverse incomplete beta.
+-- It's really hodgepodge of different approximations accumulated over years.
+--
+-- Equations are referred to by name of paper and number e.g. [AS64 2]
+-- In AS64 papers equations are not numbered so they are refered to by
+-- number of appearance starting from definition of incomplete beta.
+invIncBetaGuess beta a b p
+  -- If both a and b are less than 1 incomplete beta have inflection
+  -- point.
+  --
+  -- > x = (1 - a) / (2 - a - b)
+  --
+  -- We approximate incomplete beta by neglecting one of factors under
+  -- integral and then rescaling result of integration into [0,1]
+  -- range.
+  | a < 1 && b < 1 =
+    let x_infl = (1 - a) / (2 - a - b)
+        p_infl = incompleteBeta a b x_infl
+        x | p < p_infl = let xg = (a * p     * exp beta) ** (1/a) in xg / (1+xg)
+          | otherwise  = let xg = (b * (1-p) * exp beta) ** (1/b) in 1 - xg/(1+xg)
+    in x
+  -- If both a and b larger or equal that 1 but not too big we use
+  -- same approximation as above but calculate it a bit differently
+  | a+b <= 6 && a>1 && b>1 =
+    let x_infl = (a - 1) / (a + b - 2)
+        p_infl = incompleteBeta a b x_infl
+        x | p < p_infl = exp ((log(p * a) + beta) / a)
+          | otherwise  = 1 - exp((log((1-p) * b) + beta) / b)
+    in x
+  -- For small a and not too big b we use approximation from boost.
+  | b < 5 && a <= 1 =
+    let x | p**(1/a) < 0.5 = (p * a * exp beta) ** (1/a)
+          | otherwise      = 1 - (1 - p ** (b * exp beta))**(1/b)
+    in x
+  -- When a>>b and both are large approximation from [Temme1992],
+  -- section 4 "the incomplete gamma function case" used. In this
+  -- region it greatly improves over other approximation (AS109, AS64,
+  -- "Numerical Recipes")
+  --
+  -- FIXME: It could be used when b>>a too but it require inverse of
+  --        upper incomplete gamma to be precise enough. In current
+  --        implementation it loses precision in horrible way (40
+  --        order of magnitude off for sufficiently small p)
+  | a+b > 5 &&  a/b > 4 =
+    let -- Calculate initial approximation to eta using eq 4.1
+        eta0 = invIncompleteGamma b (1-p) / a
+        mu   = b / a            -- Eq. 4.3
+        -- A lot of helpers for calculation of
+        w    = sqrt(1 + mu)     -- Eq. 4.9
+        w_2  = w * w
+        w_3  = w_2 * w
+        w_4  = w_2 * w_2
+        w_5  = w_3 * w_2
+        w_6  = w_3 * w_3
+        w_7  = w_4 * w_3
+        w_8  = w_4 * w_4
+        w_9  = w_5 * w_4
+        w_10 = w_5 * w_5
+        d    = eta0 - mu
+        d_2  = d * d
+        d_3  = d_2 * d
+        d_4  = d_2 * d_2
+        w1   = w + 1
+        w1_2 = w1 * w1
+        w1_3 = w1 * w1_2
+        w1_4 = w1_2 * w1_2
+        -- Evaluation of eq 4.10
+        e1 = (w + 2) * (w - 1) / (3 * w)
+           + (w_3 + 9 * w_2 + 21 * w + 5) * d
+             / (36 * w_2 * w1)
+           - (w_4 - 13 * w_3 + 69 * w_2 + 167 * w + 46) * d_2
+             / (1620 * w1_2 * w_3)
+           - (7 * w_5 + 21 * w_4 + 70 * w_3 + 26 * w_2 - 93 * w - 31) * d_3
+             / (6480 * w1_3 * w_4)
+           - (75 * w_6 + 202 * w_5 + 188 * w_4 - 888 * w_3 - 1345 * w_2 + 118 * w + 138) * d_4
+             / (272160 * w1_4 * w_5)
+        e2 = (28 * w_4 + 131 * w_3 + 402 * w_2 + 581 * w + 208) * (w - 1)
+             / (1620 * w1 * w_3)
+           - (35 * w_6 - 154 * w_5 - 623 * w_4 - 1636 * w_3 - 3983 * w_2 - 3514 * w - 925) * d
+             / (12960 * w1_2 * w_4)
+           - ( 2132 * w_7 + 7915 * w_6 + 16821 * w_5 + 35066 * w_4 + 87490 * w_3
+             + 141183 * w_2 + 95993 * w + 21640
+             ) * d_2
+             / (816480 * w_5 * w1_3)
+           - ( 11053 * w_8 + 53308 * w_7 + 117010 * w_6 + 163924 * w_5 + 116188 * w_4
+             - 258428 * w_3 - 677042 * w_2 - 481940 * w - 105497
+             ) * d_3
+             / (14696640 * w1_4 * w_6)
+        e3 = -( (3592 * w_7 + 8375 * w_6 - 1323 * w_5 - 29198 * w_4 - 89578 * w_3
+                - 154413 * w_2 - 116063 * w - 29632
+                ) * (w - 1)
+              )
+              / (816480 * w_5 * w1_2)
+           - ( 442043 * w_9 + 2054169 * w_8 + 3803094 * w_7 + 3470754 * w_6 + 2141568 * w_5
+             - 2393568 * w_4 - 19904934 * w_3 - 34714674 * w_2 - 23128299 * w - 5253353
+             ) * d
+             / (146966400 * w_6 * w1_3)
+           - ( 116932 * w_10 + 819281 * w_9 + 2378172 * w_8 + 4341330 * w_7 + 6806004 * w_6
+             + 10622748 * w_5 + 18739500 * w_4 + 30651894 * w_3 + 30869976 * w_2
+             + 15431867 * w + 2919016
+             ) * d_2
+             / (146966400 * w1_4 * w_7)
+        eta = evaluatePolynomialL (1/a) [eta0, e1, e2, e3]
+        -- Now we solve eq 4.2 to recover x using Newton iterations
+        u       = eta - mu * log eta + (1 + mu) * log(1 + mu) - mu
+        cross   = 1 / (1 + mu);
+        lower   = if eta < mu then cross else 0
+        upper   = if eta < mu then 1     else cross
+        x_guess = (lower + upper) / 2
+        func x  = ( u + log x + mu*log(1 - x)
+                  , 1/x - mu/(1-x)
+                  )
+        Root x0 = newtonRaphson 1e-8 (lower, x_guess, upper) func
+    in x0
+  -- For large a and b approximation from AS109 (Carter
+  -- approximation). It's reasonably good in this region
+  | a > 1 && b > 1 =
+      let r = (y*y - 3) / 6
+          s = 1 / (2*a - 1)
+          t = 1 / (2*b - 1)
+          h = 2 / (s + t)
+          w = y * sqrt(h + r) / h - (t - s) * (r + 5/6 - 2 / (3 * h))
+      in a / (a + b * exp(2 * w))
+  -- Otherwise we revert to approximation from AS64 derived from
+  -- [AS64 2] when it's applicable.
+  --
+  -- It slightly reduces average number of iterations when `a' and
+  -- `b' have different magnitudes.
+  | chi2 > 0 && ratio > 1 = 1 - 2 / (ratio + 1)
+  -- If all else fails we use approximation from "Numerical
+  -- Recipes". It's very similar to approximations [AS64 4,5] but
+  -- it never goes out of [0,1] interval.
+  | otherwise = case () of
+      _| p < t / w  -> (a * p * w) ** (1/a)
+       | otherwise  -> 1 - (b * (1 - p) * w) ** (1/b)
+       where
+         lna = log $ a / (a+b)
+         lnb = log $ b / (a+b)
+         t   = exp( a * lna ) / a
+         u   = exp( b * lnb ) / b
+         w   = t + u
+  where
+    -- Formula [AS64 2]
+    ratio = (4*a + 2*b - 2) / chi2
+    -- Quantile of chi-squared distribution. Formula [AS64 3].
+    chi2 = 2 * b * (1 - t + y * sqrt t) ** 3
       where
-        -- Formula [2]
-        ratio = (4*a + 2*b - 2) / chi2
-        -- Quantile of chi-squared distribution. Formula [3].
-        chi2 = 2 * b * (1 - t + y * sqrt t) ** 3
-          where
-            t   = 1 / (9 * b)
-        -- `y' is Hasting's approximation of p'th quantile of standard
-        -- normal distribution.
-        y   = r - ( 2.30753 + 0.27061 * r )
-                  / ( 1.0 + ( 0.99229 + 0.04481 * r ) * r )
-          where
-            r = sqrt $ - 2 * log p
+        t   = 1 / (9 * b)
+    -- `y' is Hasting's approximation of p'th quantile of standard
+    -- normal distribution.
+    y   = r - ( 2.30753 + 0.27061 * r )
+              / ( 1.0 + ( 0.99229 + 0.04481 * r ) * r )
+      where
+        r = sqrt $ - 2 * log p
 
 
 
@@ -603,6 +758,24 @@ log1p x
               -0.10324619158271569595141333961932e-15
              ]
 
+-- | Compute log(1+x)-x:
+log1pmx :: Double -> Double
+log1pmx x
+  | x <  -1        = error "Domain error"
+  | x == -1        = m_neg_inf
+  | ax > 0.95      = log(1 + x) - x
+  | ax < m_epsilon = -(x * x) /2
+  | otherwise      = - x * x * sumPowerSeries (-x) (recip <$> enumSequenceFrom 2)
+  where
+   ax = abs x
+
+-- | Compute @exp x - 1@ without loss of accuracy for x near zero.
+expm1 :: Double -> Double
+expm1 = c_expm1
+
+foreign import ccall "expm1" c_expm1 :: Double -> Double
+
+
 
 -- | /O(log n)/ Compute the logarithm in base 2 of the given value.
 log2 :: Int -> Int
@@ -615,7 +788,9 @@ log2 v0
                                    in go (i-1) (r .|. si) (v `shiftR` si)
                 | otherwise      = go (i-1) r v
     b = U.unsafeIndex bv
-    !bv = U.fromList [0x2, 0xc, 0xf0, 0xff00, 0xffff0000, 0xffffffff00000000]
+    !bv = U.fromList [ 0x02, 0x0c, 0xf0, 0xff00
+                     , fromIntegral (0xffff0000 :: Word)
+                     , fromIntegral (0xffffffff00000000 :: Word)]
     !sv = U.fromList [1,2,4,8,16,32]
 
 
@@ -652,7 +827,9 @@ logFactorial n
 -- | Calculate the error term of the Stirling approximation.  This is
 -- only defined for non-negative values.
 --
--- > stirlingError @n@ = @log(n!) - log(sqrt(2*pi*n)*(n/e)^n)
+-- \[
+-- \operatorname{stirlingError}(n) = \log(n!) - \log(\sqrt{2\pi n}\frac{n}{e}^n)
+-- \]
 stirlingError :: Double -> Double
 stirlingError n
   | n <= 15.0   = case properFraction (n+n) of
@@ -742,9 +919,14 @@ n `choose` k
     max64          = fromIntegral (maxBound :: Int64)
     round64 x      = round x :: Int64
 
--- | Compute ψ0(/x/), the first logarithmic derivative of the gamma
--- function. Uses Algorithm AS 103 by Bernardo, based on Minka's C
--- implementation.
+-- | Compute ψ(/x/), the first logarithmic derivative of the gamma
+--   function.
+--
+-- \[
+-- \psi(x) = \frac{d}{dx} \ln \left(\Gamma(x)\right) = \frac{\Gamma'(x)}{\Gamma(x)}
+-- \]
+--
+-- Uses Algorithm AS 103 by Bernardo, based on Minka's C implementation.
 digamma :: Double -> Double
 digamma x
     | isNaN x || isInfinite x                  = m_NaN
